/*
    libtree : C++ lib that helps to work with tree-like data structures
    Copyright (C) 2019 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include    "tests.h"
#include    "tests_libtree.h"

#include    <sys/stat.h>
#include    <functional>

#define     DTYPE_XML   "xml"
#define     DTYPE_JSON  "json"

//-----------------------------------------------------------------------
// Global variables

map<string, function<bool(const string&, const int&, string &)>> unit_tests = {
    {DTYPE_XML,  TreeTest},
    {DTYPE_JSON, TreeTest}
    };
map<string, int> doc_types = {
    {DTYPE_XML,  TYPE_XML},
    {DTYPE_JSON, TYPE_JSON}
    };
map<string, string> test_files;

//-----------------------------------------------------------------------------
// Print the help note

void PrintHelp(const string &exe){

    cout << "Usage:" << endl
         << "\t" << exe << " <test_data_dir_path> [--json] [--xml] [--help]"
            << endl
         << endl
         << "JSON and XML tests are enabled by default, a particular test "
            << endl
         << "can be specified to disable another one." << endl;
    }

//-----------------------------------------------------------------------------
// Check that specified file exists

bool FileExist(const string &file_path){

    struct stat buffer;

    return stat(file_path.c_str(), &buffer) == 0;
    }

//-----------------------------------------------------------------------------
// Print given error message and terminate the execution

void Error(const string &msg, const string &preface=""){

    if(!preface.empty()){
        cout << "...(" << preface << ") ";
        }

    cout << "[ERROR";

    if(!msg.empty()){
        cout << ": " << msg;
        }

    cout << "]" << endl;

    exit(1);
    }

//-----------------------------------------------------------------------------
// Run specified unit test

void RunTest(const string &unit){

    // Printing a separator
    cout << string(80, '=') << endl;

    string err;
    const string msg = "Running " + unit + " test";

    cout << msg << "..." << endl;

    // Validating a specified name of a unit test
    if(unit_tests.find(unit) == unit_tests.end() ||
            test_files.find(unit) == test_files.end() ||
            doc_types.find(unit)  == doc_types.end()){
        Error("There is no \"" + unit + "\" unit test");
        }

    if(!unit_tests[unit](test_files[unit], doc_types[unit], err)){
        Error(err, msg);
        }

    cout << "...(" << msg << ") [OK]" << endl;
    }

//-----------------------------------------------------------------------------

int main(int argc, char *argv[]){

    string exe, test_data_dir_path;
    bool xml_test_enabled = false;
    bool json_test_enabled = false;
    bool display_help = false;

    // Parsing the arguments
    for(int i = 0; i < argc; i++){
        const string param = argv[i];
        if(i == 0){
            exe = param;
            }
        else if(param == "--xml"){
            xml_test_enabled = true;
            }
        else if(param == "--json"){
            json_test_enabled = true;
            }
        else if(param == "--help" || !test_data_dir_path.empty()){
            display_help = true;
            }
        else{
            test_data_dir_path = param;
            }
        }

    // Both tests are enabled by default
    if(!xml_test_enabled && !json_test_enabled){
        xml_test_enabled = true;
        json_test_enabled = true;
        }

    // Displaying the help message if there are insufficient or wrong params
    // or help is requested
    if(argc < 2 || display_help){
        PrintHelp(exe);
        return 1;
        }

    // Getting an absolute path of a data dir
    char *pdata_dir = realpath(test_data_dir_path.c_str(), NULL);
    if(!pdata_dir){
        Error("Could not resolve abs path for \"" + test_data_dir_path + "\"");
        }
    test_data_dir_path = pdata_dir;
    free(pdata_dir);
    if(test_data_dir_path.back() != '/'){
        test_data_dir_path += '/';
        }

    // Forming a map of test files
    for(auto &[unit, _] : unit_tests){
        // First element of a unit test pair is a name of that unit
        string test_file_path = test_data_dir_path + "test." + unit;
        if(!FileExist(test_file_path)){
            Error("File \"" + test_file_path + "\" is missing");
            }
        test_files[unit] = test_file_path;
        }

    // Running the tests
    for(auto &[unit, _] : unit_tests){
        if((unit == DTYPE_XML && xml_test_enabled) ||
                (unit == DTYPE_JSON && json_test_enabled)){
            RunTest(unit);
            }
        }

    return 0;
    }

//-----------------------------------------------------------------------------
