/*
    libtree : C++ lib that helps to work with tree-like data structures
    Copyright (C) 2019 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include    "tests_libtree.h"

#include    "xml.h"
#include    "json.h"

// BASE  is root >> level_1
// BASE1 is BASE >> level_20
// BASE2 is BASE >> level_21
#define     N_BRANCHES      7
#define     D_MASTER        0   // BASE1 >> (a >> b | array) >> array >> d
#define     A_SLAVE_OF_D    1   // BASE1 >> (a > tag | a)
#define     B_SLAVE_OF_D    2   // BASE1 >> (a | array) >> b > (tag |)
#define     C_SLAVE_OF_D    3   // BASE1 >> (a >> b | array) >> array >> c
#define     E_SLAVE_OF_D    4   // BASE1 >> (a >> b | array) >> array >> e
#define     X_MASTER        5   // BASE2 >> (x > tag) | x
#define     Y_SLAVE_OF_X    6   // BASE2 >> y

const array<int, 2> doc_types = {TYPE_XML, TYPE_JSON};

//-----------------------------------------------------------------------------

pair<string, vector<tree::TBranchDraft>> FormBranchDrafts(const int &doc_type,
        string &err){

    string root;
    tree::TBranchDraft branch_draft;
    vector<tree::TBranchDraft> branch_drafts;

    branch_drafts.resize(N_BRANCHES);

    #define ROOT_NODE  "root"
    #define BASE_NODES {"level_1"}

    root = ROOT_NODE;

    for(size_t i = 0; i < N_BRANCHES; i++){
        int master_branch = -1;
        bool enable_indexing = false;
        branch_draft.Add(BASE_NODES);
        // root >> level_1 >> (level20 | level 21)
        if(i < X_MASTER){
            branch_draft.Add({"level_20"});
            }
        else{
            branch_draft.Add({"level_21"});
            }
        if(i == D_MASTER || i == C_SLAVE_OF_D || i == E_SLAVE_OF_D){
            // XML:  (..) >> a >> b >> array >> (c|d|e)
            // JSON: (..) >> array >> array >> (c|d|e)
            if(doc_type == TYPE_XML){
                branch_draft.Add({"a", "b"});
                }
            else{
                branch_draft.Add({"array"});
                }
            branch_draft.Add({"array"});
            }
        switch(i){
            case D_MASTER:
                branch_draft.Add({"d"});
                break;
            case A_SLAVE_OF_D:
                // XML:  (..) >> a > tag
                // JSON: (..) >> a
                if(doc_type == TYPE_XML){
                    branch_draft.Add({"a"}, "tag");
                    }
                else{
                    branch_draft.Add({"a"});
                    }
                break;
            case B_SLAVE_OF_D:
                // XML:  (..) >> a >> b > tag
                // JSON: (..) >> array >> b
                if(doc_type == TYPE_XML){
                    branch_draft.Add({"a", "b"}, "tag");
                    }
                else{
                    branch_draft.Add({"array", "b"});
                    }
                break;
            case C_SLAVE_OF_D:
                branch_draft.Add({"c"});
                break;
            case E_SLAVE_OF_D:
                branch_draft.Add({"e"});
                break;
            case X_MASTER:
                // XML:  (..) >> x > tag
                // JSON: (..) >> x
                if(doc_type == TYPE_XML){
                    branch_draft.Add({"x"}, "tag");
                    }
                else{
                    branch_draft.Add({"x"});
                    }
                break;
            case Y_SLAVE_OF_X:
                // XML:  (..) >> x >> y
                // JSON: (..) >> y
                if(doc_type == TYPE_XML){
                    branch_draft.Add({"x"});
                    }
                branch_draft.Add({"y"});
                break;
            default:
                err = "No case for branch mark '" + to_string(i) + "'";
                return {};
            }
        if(i != D_MASTER && i != X_MASTER){
            master_branch = (i > X_MASTER) ? X_MASTER : D_MASTER;
            }
        branch_draft.Complete(i, master_branch, enable_indexing);
        branch_drafts[i] = branch_draft;
        branch_draft.Clear();
        }

    #undef ROOT_NODE
    #undef BASE_NODES

    return {root, branch_drafts};
    }

//-----------------------------------------------------------------------------

bool CheckResults(const tree::TSelectedData &selected_data, const int &doc_type,
        string &err){

    for(size_t i = 0; i <= 1; i++){
        if(i >= selected_data.size()){
            err = "Not enough entries";
            return false;
            }
        for(size_t j = 0; j < N_BRANCHES; j++){
            if(j >= selected_data.at(i).size()){
                err = "Not enough items in entry №" + to_string(i);
                return false;
                }
            string entry_txt = selected_data[i][j].Stringify();
            string sample;
            switch(j){
                case D_MASTER:
                    sample =
                        "{val_d" + to_string(i) + "00},{val_d" + to_string(i) + "01},"
                        "{val_d" + to_string(i) + "10},{val_d" + to_string(i) + "11}";
                    break;
                case A_SLAVE_OF_D:
                    sample =
                        "{val_a" + to_string(i) + "},{val_a" + to_string(i) + "},"
                        "{val_a" + to_string(i) + "},{val_a" + to_string(i) + "}";
                    break;
                case B_SLAVE_OF_D:
                    sample =
                        "{val_b" + to_string(i) + "0},{val_b" + to_string(i) + "0},"
                        "{val_b" + to_string(i) + "1},{val_b" + to_string(i) + "1}";
                    break;
                case C_SLAVE_OF_D:
                    sample = "{1},{0},{1},{0}";
                    break;
                case E_SLAVE_OF_D:
                    sample =
                        "{val_e" + to_string(i) + "00},{val_e" + to_string(i) + "01},"
                        "{val_e" + to_string(i) + "10},{val_e" + to_string(i) + "11}";
                    break;
                case X_MASTER:
                    sample =
                        "{val_x" + to_string(i) + "0},"
                        "{val_x" + to_string(i) + "1}";
                    break;
                case Y_SLAVE_OF_X:
                    // One of the two things that XML component can't do is
                    // to write an array of values into a single item at once
                    // (the second thing is handling of a boolean values)
                    for(int k = 0; k < 2; k++){
                        const string s = "{"
                            "val_y" + to_string(i) + to_string(k % 2) + "0,"
                            "val_y" + to_string(i) + to_string(k % 2) + "1}";
                        sample += (doc_type == TYPE_JSON) ? "{" + s + "}" : s;
                        if(k == 0){
                            sample += ",";
                            }
                        }
                    break;
                default:
                    err = "Wrong case";
                    return false;
                }
            if(entry_txt != sample){
                err = "Wrong selection, entry #" + to_string(i) + " "
                    "item #" + to_string(j) + "; "
                    "Expected \"" + sample + "\", got \"" + entry_txt + "\"";
                return false;
                }
            }
        }

    return true;
    }

//-----------------------------------------------------------------------------

bool TreeTest(const string &file_path, const int &doc_type, string &err){

    tree::TTree *tree = NULL;
    string root;
    vector<tree::TBranchDraft> branch_drafts;

    // This flag will be switched at the end if this test will get there
    bool success = false;

    // Checking given doc type
    if(find(doc_types.begin(), doc_types.end(), doc_type) == doc_types.end()){
        err = "Wrong doc type index";
        goto TreeTestReturn;
        }

    // Forming the draft
    tie(root, branch_drafts) = FormBranchDrafts(doc_type, err);
    if(branch_drafts.empty()){
        goto TreeTestReturn;
        }

    // Creating tree object of needed subtype
    switch(doc_type){
        case TYPE_XML:
            tree = new tree::TXmlTree();
            break;
        case TYPE_JSON:
            tree = new tree::TJsonTree();
            break;
        }

    // Checking that a TTree object is created
    if(!tree){
        err = "Unable to spawn a TTree object";
        goto TreeTestReturn;
        }

    // Constructing a tree of needed nodes
    if(!tree->Construct(root, branch_drafts)){
        err = "Unable to construct the tree";
        goto TreeTestReturn;
        }

    // Extracting the data from a test XML-file
    if(!tree->Select(file_path, err)){
        goto TreeTestReturn;
        }

    // Checking the results
    if(!CheckResults(tree->SelectedData(), doc_type, err)){
        goto TreeTestReturn;
        }

    // Test OK
    success = true;

    TreeTestReturn:
    if(tree){
        delete tree;
        }

    return success;
    }

//-----------------------------------------------------------------------------
